open C

exception NotImplemented
exception UndefinedSemantics

module type DOMAIN =
	sig
		type elt (* the type of abstract domain elements *) 
		val bot: elt
		val join: elt -> elt -> elt (* least upper bound *)
		val leq: elt -> elt -> bool (* less than or equal to *)
		val string_of_elt: elt -> string
		val add: elt -> elt -> elt (* addition between elements *)
		val sub: elt -> elt -> elt (* subtraction between elements *)
		val mul: elt -> elt -> elt (* multiplication between elements *)
	end

module type INTSET_DOMAIN =
	sig
		include DOMAIN
		val filter: (int -> bool) -> elt -> elt
		val make: int list -> elt
	end

module type PARITY_DOMAIN =
	sig
		include DOMAIN
		val top: elt
		val meet: elt -> elt -> elt (* greatest lower bound *)
		val make: int -> elt
	end
	
module type INTERVAL_DOMAIN =
	sig
		include DOMAIN
		type bound = Z of int | Pinfty | Ninfty
		val top: elt
		val meet: elt -> elt -> elt (* greatest lower bound *)
		val make: bound -> bound -> elt
		val widen: elt -> elt -> elt 
		val narrow: elt -> elt -> elt 
	end
	
module IntCPO = 
struct
	type elt = Bot | Int of int
	let bot = Bot
	let join t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> t2 
		| _, Bot -> t1
		| _ -> raise (Failure "can't join two different integers!")
	
	let of_int n = Int n 
	
	let leq t1 t2 =
		match t1, t2 with 
		| Bot, _ -> true 
		| _, Bot -> false
		| (Int n1, Int n2) -> n1 <= n2   

	let string_of_elt t =
		match t with 
		| Bot -> "bot"
		| Int n -> string_of_int n 

	let add t1 t2 =
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Int n1, Int n2 -> Int (n1 + n2) 
	
	let sub t1 t2 =
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Int n1, Int n2 -> Int (n1 - n2)

	let mul t1 t2 =
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Int n1, Int n2 -> Int (n1 * n2)
end

(* helper function for IntSet *)
(* return the cartesian product of alst and blst *)
let rec cartesian alst blst =
  let rec cartesian_sub a blst =
      match blst with [] -> [] | hd :: tl -> (a, hd) :: (cartesian_sub a tl)
  in
  match alst with
  | hd::tl -> (cartesian_sub hd blst) @ (cartesian tl blst)
  | [] -> []
	
module IntsetCPO : INTSET_DOMAIN = 
struct
	(* A data structure for integer sets *)
	(* see https://ocaml.org/api/Set.Make.html *)
	include Set.Make (struct 
		type t = int
		let compare = Pervasives.compare 
	end)
	type elt = t 
	let make ns = of_list ns 
	let bot = empty
	let join = union
	let leq = subset
	let apply_op_cartesian op lst1 lst2 = 
		List.fold_left (fun acc (v1, v2) ->
			add (op v1 v2) acc  
		) empty (cartesian lst1 lst2)
	let add t1 t2 = apply_op_cartesian (+) (elements t1) (elements t2)
	let sub t1 t2 = apply_op_cartesian (-) (elements t1) (elements t2)
	let mul t1 t2 = apply_op_cartesian ( * ) (elements t1) (elements t2)
	let string_of_elt t = 
		(fold (fun n str -> Printf.sprintf "%s, %d" str n) t "{") ^ "}"  		 
end


module ParityCPO : PARITY_DOMAIN = struct
	type elt = Even | Odd | Top | Bot
	let top = Top
	let bot = Bot 
	let make n = if (n mod 2) = 0 then Even else Odd
	let join t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> t2
		| _, Bot -> t1 
		| Top, _ -> Top
		| _, Top -> Top
		| _ -> if t1 = t2 then t1 else Top
	
	let meet t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot 
		| Top, _ -> t2
		| _, Top -> t1
		| _ -> if t1 = t2 then t1 else Bot
	
	let leq t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> true 
		| _, Top -> true 
		| _ -> (t1 = t2)
	
	let string_of_elt t = 
		match t with 
		| Bot -> "bot_parity"
		| Top -> "top_parity"
		| Even -> "even"
		| Odd -> "odd"	

	let add t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Top, _ -> Top
		| _, Top -> Top 
		| _ -> if t1 = t2 then Even else Odd 
	
	let sub = add
	
	let mul t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Top, _ -> Top
		| _, Top -> Top 
		| Even, Odd -> Even
		| Odd, Even -> Even
		| Even, Even -> Even   
		| Odd, Odd -> Odd
end


module IntervalCPO : INTERVAL_DOMAIN = 
struct
	type bound = Z of int | Pinfty | Ninfty
	type elt = Bot | Intv of (bound * bound)
	let top = Intv (Ninfty, Pinfty)
	let bot = Bot 
	let make lb ub = Intv (lb, ub)
  let ge b1 b2 = 
		match (b1, b2) with
    | (Pinfty, Pinfty) -> true
    | (Pinfty, _) -> true
    | (_, Pinfty) -> false
    | (Ninfty, Ninfty) -> true
    | (Ninfty, _) -> false
    | (_, Ninfty) -> true
    | (Z n1, Z n2) -> n1 >= n2

  let le b1 b2 = 
		match (b1, b2) with
		| (Pinfty, Pinfty) -> true
		| (Pinfty, _) -> false
		| (_, Pinfty) -> true
		| (Ninfty, Ninfty) -> true
		| (Ninfty, _) -> true
		| (_, Ninfty) -> false
		| (Z n1, Z n2) -> n1 <= n2

	let lt b1 b2 = (le b1 b2) && (not (le b2 b1))
	let order_interval (i11, i12) (i21, i22) = (ge i11 i21) && (le i12 i22)
	
	let max_bounds bounds = 
		List.fold_left (fun acc bound -> if ge bound acc then bound else acc) Ninfty bounds  
	let min_bounds bounds = 
		List.fold_left (fun acc bound -> if le bound acc then bound else acc) Pinfty bounds
	
  let leq t1 t2 = 
		match (t1, t2) with
    | (Bot, _) -> true
    | (_, Bot) -> false
    | (Intv i1), (Intv i2) -> order_interval i1 i2

	let join t1 t2 = 
		match (t1, t2) with
    | (Bot, _) -> t2
    | (_, Bot) -> t1
    | (Intv i1), (Intv i2) ->
			let (i11, i12) = i1 in
			let (i21, i22) = i2 in  
			Intv ((min_bounds [i11; i21]), (max_bounds [i12; i22]))

	let meet t1 t2 = 
		match (t1, t2) with
    | (Bot, _) -> Bot
    | (_, Bot) -> Bot
    | (Intv i1), (Intv i2) ->
			let (i11, i12) = i1 in
			let (i21, i22) = i2 in
			let lb, ub = (max_bounds [i11; i21]), (min_bounds [i12; i22]) in
			if (le ub lb) && not (le lb ub) then Bot    
			else Intv (lb, ub)
			
	let widen t1 t2 = 
		match (t1, t2) with 
		| (Bot, _) -> t2
    | (_, Bot) -> t1
    | (Intv i1), (Intv i2) ->
			let (i11, i12) = i1 in
			let (i21, i22) = i2 in
			let lb = if lt i21 i11 then Ninfty else i11 in  
			let ub = if lt i12 i22 then Pinfty else i12 in 
			Intv (lb, ub) 		
  
	let narrow t1 t2 = 
		match (t1, t2) with 
		| (Bot, _) -> Bot
    | (_, Bot) -> Bot
    | (Intv i1), (Intv i2) ->
			let (i11, i12) = i1 in
			let (i21, i22) = i2 in
			let lb = if i11 = Ninfty then i21 else i11 in  
			let ub = if i12 = Pinfty then i22 else i12 in 
			Intv (lb, ub) 		
			
	let string_of_bound i = 
		match i with
    | Pinfty -> "oo"
		| Ninfty -> "-oo"
		| Z n -> string_of_int n
		
	let string_of_elt = function
    Bot -> "bot_interval"
  	| Intv (i1, i2) -> "["^(string_of_bound i1)^", "^(string_of_bound i2)^"]"

  let add_interval i1 i2 =
		let (++) i1 i2 = 
			match (i1, i2) with
				(Ninfty, _) -> Ninfty
  		| (_, Ninfty) -> Ninfty
  		| (Pinfty, _) -> Pinfty
  		| (_, Pinfty) -> Pinfty
  		| (Z n1, Z n2) -> Z (n1+n2)
		in
    match (i1, i2) with
    ( (i11, i12) , (i21, i22) ) -> ( i11 ++ i21, i12 ++ i22)
  
	let add t1 t2 = 
		match (t1, t2) with
			(Bot, _) -> Bot
    | (_, Bot) -> Bot
    | (Intv (i11, i12), Intv (i21, i22)) -> Intv (add_interval (i11, i12) (i21, i22))

  let neg_interval (i1, i2) =
    let minus i1 = 
			match i1 with
      	Ninfty -> Pinfty
    	| Pinfty -> Ninfty
    	| Z n1 -> Z (-n1)
		in
		(minus i2, minus i1)
  
	let neg t =  
		match t with
			(Intv i) -> Intv (neg_interval i)
    | Bot -> Bot

	let sub t1 t2 = add t1 (neg t2)
		
	let mul_interval i1 i2 =
		let mult_bound i1 i2 = 
			match (i1, i2) with
			| (Ninfty, Pinfty) -> Ninfty
			| (Pinfty, Ninfty) -> Ninfty
			|	(Ninfty, _) -> Ninfty
  		| (_, Ninfty) -> Ninfty
  		| (Pinfty, _) -> Pinfty
  		| (_, Pinfty) -> Pinfty
  		| (Z n1, Z n2) -> Z (n1 * n2)
		in
    match (i1, i2) with
    ( (i11, i12) , (i21, i22) ) ->
			let all_bounds = [(mult_bound i11 i21);(mult_bound i11 i22);(mult_bound i12 i21);(mult_bound i12 i22)] in  
			(min_bounds all_bounds, max_bounds all_bounds) 
		
	let mul t1 t2 = 
		match (t1, t2) with
			(Bot, _) -> Bot
    | (_, Bot) -> Bot
    | (Intv (i11, i12), Intv (i21, i22)) -> Intv (mul_interval (i11, i12) (i21, i22))
end


module VarMap = Map.Make (struct 
		type t = var
		let compare = String.compare
	end)
	 
module MakeMemCPO (D : DOMAIN) = 
struct
	(* see https://ocaml.org/api/Map.Make.html *)
	include VarMap
	
	type t = D.elt VarMap.t (* type : var -> D.elt *)

	let val_compare t1 t2 = 
		if (D.leq t1 t2) && (D.leq t2 t1) then 0
		else if (D.leq t1 t2) then -1 
		else 1
			
	let find x t = 
		try 
			find x t 
		with Not_found -> D.bot
	let keys m = fold (fun k _ acc -> k :: acc) m []
	
	(* join two memories *)
  let join m1 m2 = 
  	let vars = (keys m1) @ (keys m2) in
  	List.fold_left (fun m var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		add var (D.join v1 v2) m   
  	) empty vars	      
  
	(* widen two memories *)
  let widen widen_op m1 m2 = 
  	let vars = (keys m1) @ (keys m2) in
  	List.fold_left (fun m var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		add var (widen_op v1 v2) m   
  	) empty vars	      
	
	(* narrow two memories *)
  let narrow narrow_op m1 m2 = 
  	let vars = (keys m1) @ (keys m2) in
  	List.fold_left (fun m var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		add var (narrow_op v1 v2) m   
  	) empty vars	 
				
  (* check if m1 <= m2 for two memories m1 m2 *)
  let leq m1 m2 = 
  	let vars = (keys m1) in
  	List.fold_left (fun acc var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		acc && (D.leq v1 v2)
  	) true vars	      
	
	let eq m1 m2 = (leq m1 m2) && (leq m2 m1) 	
  
  let string_of_t mem = 
  	fold (fun x v str -> 
			Printf.sprintf "%s\n%s -> %s" str x (D.string_of_elt v)
		) mem ""
end 
	
(* Data structure for various kinds of memories *)
module Mem = MakeMemCPO(IntCPO) 
module IntsetMem = MakeMemCPO(IntsetCPO)
module ParityMem = MakeMemCPO(ParityCPO)
module IntervalMem = MakeMemCPO(IntervalCPO)
