open C
open Domains
	
let rec eval_exp : exp -> Mem.t -> IntCPO.elt 
= fun exp mem -> 
	match exp with 
	| CONST n -> IntCPO.of_int n 
  | VAR x -> Mem.find x mem 
  | ADD (e1, e2) -> IntCPO.add (eval_exp e1 mem) (eval_exp e2 mem)  
  | SUB (e1, e2) -> IntCPO.sub (eval_exp e1 mem) (eval_exp e2 mem)
  | MUL (e1, e2) -> IntCPO.mul (eval_exp e1 mem) (eval_exp e2 mem) 

and eval_cond: cond -> Mem.t -> bool
= fun cond mem -> 
	match cond with 
	| TRUE -> true
	| FALSE -> false
	| LE (x, n) ->
		let v = Mem.find x mem in  
		IntCPO.leq v (IntCPO.of_int n)
	| EQ (x, n) -> 
		let v = Mem.find x mem in  
		(IntCPO.leq v (IntCPO.of_int n)) && (IntCPO.leq (IntCPO.of_int n) v) 
	| GT (x, n) -> 
		let v = Mem.find x mem in  
		(not (IntCPO.leq v (IntCPO.of_int n))) && (IntCPO.leq (IntCPO.of_int n) v)
	| NEQ (x, n) -> eval_cond (NOT (EQ (x, n))) mem 
	| NOT cond' -> not (eval_cond cond' mem)

let rec interpret : program -> Mem.t -> Mem.t
= fun pgm mem -> 
	(* raise NotImplemented *)
	(* print_endline (string_of_cmd pgm);   *)
	(* print_endline (Mem.string_of_t mem); *)
	match pgm with 
	| SKIP -> mem 
	| IF (cond, c1, c2) -> 
		if (eval_cond cond mem) then interpret c1 mem else interpret c2 mem  
	| WHILE (cond, c) -> 
		if (eval_cond cond mem) then
			interpret (SEQ(c, WHILE (cond, c))) mem
		else mem 
	| ASSIGN (x, e) -> Mem.add x (eval_exp e mem) mem 
	| READ x -> Mem.add x (IntCPO.of_int (read_int ())) mem 
	| SEQ (c1, c2) -> interpret c2 (interpret c1 mem)
	| PRINT e -> (print_endline (IntCPO.string_of_elt (eval_exp e mem)); mem)  
	
	
let rec eval_exp_intset : exp -> IntsetMem.t -> IntsetCPO.elt 
= fun exp mem ->
	match exp with 
	| CONST n -> IntsetCPO.make [n] 
  | VAR x -> IntsetMem.find x mem  
  | ADD (e1, e2) -> 
		let v1 = (eval_exp_intset e1 mem) in
		let v2 = (eval_exp_intset e2 mem) in 
		IntsetCPO.add v1 v2
  | SUB (e1, e2) -> 
		let v1 = (eval_exp_intset e1 mem) in
		let v2 = (eval_exp_intset e2 mem) in 
		IntsetCPO.sub v1 v2
  | MUL (e1, e2) -> 
		let v1 = (eval_exp_intset e1 mem) in
		let v2 = (eval_exp_intset e2 mem) in 
		IntsetCPO.mul v1 v2 						
	
let rec filter_mem_intset : cond -> IntsetMem.t -> IntsetMem.t
= fun cond mem ->
	match cond with 
	| TRUE -> mem
	| FALSE -> IntsetMem.empty 
	| LE (x, n) -> 
		let vs = IntsetMem.find x mem in 
		IntsetMem.add x (IntsetCPO.filter (fun x -> x <= n) vs) mem
	| EQ (x, n) -> 
		let vs = IntsetMem.find x mem in 
		IntsetMem.add x (IntsetCPO.filter (fun x -> x = n) vs) mem
	| GT (x, n) -> 
		let vs = IntsetMem.find x mem in 
		IntsetMem.add x (IntsetCPO.filter (fun x -> x > n) vs) mem
	| NEQ (x, n) -> 
		let vs = IntsetMem.find x mem in 
		IntsetMem.add x (IntsetCPO.filter (fun x -> x <> n) vs) mem
	| NOT cond -> 
		let cond' = 
  		(match cond with 
  		| LE (x, n) -> GT (x, n)
			| EQ (x, n) -> NEQ (x, n)
			| GT (x, n) -> LE (x, n)
			| NEQ (x, n) -> EQ (x, n)
			| NOT cond -> cond 
			| TRUE -> FALSE
			| FALSE -> TRUE
  		)
		in
		filter_mem_intset cond' mem


let rec interpret_collect : program -> IntsetMem.t -> IntsetMem.t 
= fun pgm mem ->
	(* raise NotImplemented *) 
	match pgm with 
	| SKIP -> mem 
	| IF (cond, c1, c2) -> 
		let mem1 = interpret_collect c1 (filter_mem_intset cond mem) in 
		let mem2 = interpret_collect c2 (filter_mem_intset (NOT cond) mem) in 
		IntsetMem.join mem1 mem2  
	| WHILE (cond, c) -> 
		let rec iter init_mem mem = 
			let mem' =
				let run_loop_body =  
					interpret_collect c (filter_mem_intset cond mem)
				in
				IntsetMem.join init_mem run_loop_body 
			in
			if IntsetMem.leq mem' mem then mem
			else iter init_mem mem'
		in
		filter_mem_intset (NOT cond) (iter mem mem)
	| ASSIGN (x, e) -> IntsetMem.add x (eval_exp_intset e mem) mem 
	| READ x -> IntsetMem.add x (IntsetCPO.make read_range) mem 
	| SEQ (c1, c2) -> interpret_collect c2 (interpret_collect c1 mem)	
	| _ -> mem


let rec eval_exp_parity : exp -> ParityMem.t -> ParityCPO.elt 
= fun exp mem ->
	match exp with 
	| CONST n -> ParityCPO.make n 
  | VAR x -> ParityMem.find x mem  
  | ADD (e1, e2) -> 
		let v1 = (eval_exp_parity e1 mem) in
		let v2 = (eval_exp_parity e2 mem) in 
		ParityCPO.add v1 v2
  | SUB (e1, e2) -> 
		let v1 = (eval_exp_parity e1 mem) in
		let v2 = (eval_exp_parity e2 mem) in 
		ParityCPO.sub v1 v2
  | MUL (e1, e2) -> 
		let v1 = (eval_exp_parity e1 mem) in
		let v2 = (eval_exp_parity e2 mem) in 
		ParityCPO.mul v1 v2 						
	
let rec filter_mem_parity : cond -> ParityMem.t -> ParityMem.t
= fun cond mem ->
	match cond with 
	| TRUE -> mem
	| FALSE -> ParityMem.empty 
	| LE (x, n) -> mem
	| EQ (x, n) -> 
		let v = ParityMem.find x mem in 
		ParityMem.add x (ParityCPO.meet (ParityCPO.make n) v) mem
	| GT (x, n) -> mem
	| NEQ (x, n) -> mem
	| NOT cond -> 
		let cond' = 
  		(match cond with 
  		| LE (x, n) -> GT (x, n)
			| EQ (x, n) -> NEQ (x, n)
			| GT (x, n) -> LE (x, n)
			| NEQ (x, n) -> EQ (x, n)
			| NOT cond -> cond 
			| TRUE -> FALSE
			| FALSE -> TRUE
  		)
		in
		filter_mem_parity cond' mem
		
let rec interpret_parity : program -> ParityMem.t -> ParityMem.t
	= fun pgm mem -> 
	(* raise NotImplemented *)
	match pgm with 
	| SKIP -> mem 
	| IF (cond, c1, c2) -> 
		let mem1 = interpret_parity c1 (filter_mem_parity cond mem) in 
		let mem2 = interpret_parity c2 (filter_mem_parity (NOT cond) mem) in 
		ParityMem.join mem1 mem2  
	| WHILE (cond, c) -> 
		let rec iter init_mem mem = 
			let mem' =
				let run_loop_body =  
					interpret_parity c (filter_mem_parity cond mem)
				in
				ParityMem.join init_mem run_loop_body 
			in
			if ParityMem.leq mem' mem then mem
			else iter init_mem mem'
		in
		filter_mem_parity (NOT cond) (iter mem mem)
	| ASSIGN (x, e) -> ParityMem.add x (eval_exp_parity e mem) mem 
	| READ x -> 
		let abs_read_range = 
			List.fold_left (fun acc n -> 
				ParityCPO.join acc (ParityCPO.make n)
			) ParityCPO.bot read_range 
		in  
		ParityMem.add x abs_read_range mem 
	| SEQ (c1, c2) -> interpret_parity c2 (interpret_parity c1 mem)	
	| _ -> mem
	

let rec eval_exp_interval : exp -> IntervalMem.t -> IntervalCPO.elt 
= fun exp mem ->
	match exp with 
	| CONST n -> IntervalCPO.make (IntervalCPO.Z n) (IntervalCPO.Z n) 
  | VAR x -> IntervalMem.find x mem  
  | ADD (e1, e2) -> 
		let v1 = (eval_exp_interval e1 mem) in
		let v2 = (eval_exp_interval e2 mem) in 
		IntervalCPO.add v1 v2
  | SUB (e1, e2) -> 
		let v1 = (eval_exp_interval e1 mem) in
		let v2 = (eval_exp_interval e2 mem) in 
		IntervalCPO.sub v1 v2
  | MUL (e1, e2) -> 
		let v1 = (eval_exp_interval e1 mem) in
		let v2 = (eval_exp_interval e2 mem) in 
		IntervalCPO.mul v1 v2 						
	
let rec filter_mem_interval : cond -> IntervalMem.t -> IntervalMem.t
= fun cond mem ->
	match cond with 
	| TRUE -> mem
	| FALSE -> IntervalMem.empty 
	| LE (x, n) -> 
		let v = IntervalMem.find x mem in 
		IntervalMem.add x (IntervalCPO.meet (IntervalCPO.make IntervalCPO.Ninfty (IntervalCPO.Z n)) v) mem
	| EQ (x, n) -> 
		let v = IntervalMem.find x mem in 
		IntervalMem.add x (IntervalCPO.meet (IntervalCPO.make (IntervalCPO.Z n) (IntervalCPO.Z n)) v) mem
	| GT (x, n) -> 
		let v = IntervalMem.find x mem in 
		IntervalMem.add x (IntervalCPO.meet (IntervalCPO.make (IntervalCPO.Z (n+1)) IntervalCPO.Pinfty) v) mem
	| NEQ (x, n) -> 
		let mem1 = filter_mem_interval (GT (x, n)) mem in 
		let mem2 = filter_mem_interval (LE(x, n-1)) mem in 
		IntervalMem.join mem1 mem2
	| NOT cond -> 
		let cond' = 
  		(match cond with 
  		| LE (x, n) -> GT (x, n)
			| EQ (x, n) -> NEQ (x, n)
			| GT (x, n) -> LE (x, n)
			| NEQ (x, n) -> EQ (x, n)
			| NOT cond -> cond 
			| TRUE -> FALSE
			| FALSE -> TRUE
  		)
		in
		filter_mem_interval cond' mem		
	
let rec interpret_interval : program -> IntervalMem.t -> IntervalMem.t 
= fun pgm mem -> 
	(* raise NotImplemented *)
	match pgm with 
	| SKIP -> mem 
	| IF (cond, c1, c2) -> 
		let mem1 = interpret_interval c1 (filter_mem_interval cond mem) in 
		let mem2 = interpret_interval c2 (filter_mem_interval (NOT cond) mem) in 
		IntervalMem.join mem1 mem2  
	| WHILE (cond, c) -> 
		let rec widening_iter init_mem mem = 
			let mem' =
				let run_loop_body =  
					interpret_interval c (filter_mem_interval cond mem)
				in
				IntervalMem.widen IntervalCPO.widen init_mem run_loop_body 
			in
			if IntervalMem.eq mem' mem then mem
			else widening_iter init_mem mem'
		in
		let rec narrowing_iter mem = 
			let mem' =
				let run_loop_body =  
					interpret_interval c (filter_mem_interval cond mem)
				in
				IntervalMem.narrow IntervalCPO.narrow mem run_loop_body 
			in
			if IntervalMem.eq mem' mem then mem
			else narrowing_iter mem'
		in
		filter_mem_interval (NOT cond) (narrowing_iter (widening_iter mem mem))
	| ASSIGN (x, e) -> IntervalMem.add x (eval_exp_interval e mem) mem 
	| READ x ->
		let abs_read_range = 
			List.fold_left (fun acc n -> 
				IntervalCPO.join acc (IntervalCPO.make (IntervalCPO.Z n) (IntervalCPO.Z n))
			) IntervalCPO.bot read_range 
		in  
		IntervalMem.add x abs_read_range mem 
	| SEQ (c1, c2) -> interpret_interval c2 (interpret_interval c1 mem)	
	| _ -> mem
